import copy
from typing import Optional, Tuple, Union
import torch
from torch import nn
from torch.utils import checkpoint

from transformers.models.t5.modeling_t5 import (
    T5PreTrainedModel,
    T5Config,
    T5Stack,
    T5_START_DOCSTRING,
    T5_ENCODER_INPUTS_DOCSTRING,
    PARALLELIZE_DOCSTRING,
    DEPARALLELIZE_DOCSTRING,
    _CONFIG_FOR_DOC
)
from transformers.modeling_outputs import TokenClassifierOutput
from transformers.utils import (
    add_start_docstrings,
    add_start_docstrings_to_model_forward,
    replace_return_docstrings
)
from transformers.utils.model_parallel_utils import assert_device_map, get_device_map
from transformers.utils.logging import get_logger
logger = get_logger("transformers")

@add_start_docstrings(
    """T5 Model with a token classification head on top (a linear layer on top of the hidden-states output) e.g. for
    Named-Entity-Recognition (NER) tasks.
    """, 
    T5_START_DOCSTRING
)
class T5ForTokenClassification(T5PreTrainedModel):
    _keys_to_ignore_on_load_missing = [r"encoder.embed_tokens.weight"]

    def __init__(self, config: T5Config, num_labels: int = None):
        super().__init__(config)
        self.model_dim = config.d_model

        self.shared = nn.Embedding(config.vocab_size, config.d_model)

        encoder_config = copy.deepcopy(config)
        encoder_config.is_decoder = False
        encoder_config.is_encoder_decoder = False
        encoder_config.use_cache = False
        self.encoder = T5Stack(encoder_config, self.shared)

        classifier_dropout = (
            config.classifier_dropout if hasattr(config, 'classifier_dropout') else config.dropout_rate
        )
        self.dropout = nn.Dropout(classifier_dropout)
        # Use custom num_labels if provided, otherwise use config.num_labels
        self.num_labels = num_labels if num_labels is not None else config.num_labels        
        self.classifier = nn.Linear(config.d_model, self.num_labels)

        # Initialize weights and apply final processing
        self.post_init()

        # Model parallel
        self.model_parallel = False
        self.device_map = None
            
    @add_start_docstrings(PARALLELIZE_DOCSTRING)
    def parallelize(self, device_map=None):
        self.device_map = (
            get_device_map(len(self.encoder.block), range(torch.cuda.device_count()))
            if device_map is None
            else device_map
        )
        assert_device_map(self.device_map, len(self.encoder.block))
        self.encoder.parallelize(self.device_map)
        self.classifier.to(self.encoder.first_device)
        self.model_parallel = True

    @add_start_docstrings(DEPARALLELIZE_DOCSTRING)
    def deparallelize(self):
        self.encoder.deparallelize()
        self.encoder = self.encoder.to("cpu")
        self.classifier = self.classifier.to("cpu")
        self.model_parallel = False
        self.device_map = None
        torch.cuda.empty_cache()

    def get_input_embeddings(self):
        return self.shared

    def set_input_embeddings(self, new_embeddings):
        self.shared = new_embeddings
        self.encoder.set_input_embeddings(new_embeddings)

    def get_encoder(self):
        return self.encoder

    def _prune_heads(self, heads_to_prune):
        """
        Prunes heads of the model. heads_to_prune: dict of {layer_num: list of heads to prune in this layer} See base
        class PreTrainedModel
        """
        for layer, heads in heads_to_prune.items():
            self.encoder.block[layer].layer[0].SelfAttention.prune_heads(heads)

    def check_for_nans(self, tensor, name="Tensor"):
        if torch.isnan(tensor).any():
            print(f"NaNs detected in {name}: {tensor}")

    @add_start_docstrings_to_model_forward(T5_ENCODER_INPUTS_DOCSTRING)
    @replace_return_docstrings(output_type=TokenClassifierOutput, config_class=_CONFIG_FOR_DOC)

    def forward(
        self,
        input_ids: Optional[torch.LongTensor] = None,
        attention_mask: Optional[torch.FloatTensor] = None,
        head_mask: Optional[torch.FloatTensor] = None,
        inputs_embeds: Optional[torch.FloatTensor] = None,
        labels: Optional[torch.LongTensor] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = None,
    ) -> Union[Tuple[torch.FloatTensor], TokenClassifierOutput]:
        r"""
        labels (`torch.LongTensor` of shape `(batch_size, sequence_length)`, *optional*):
            Labels for computing the token classification loss. Indices should be in `[0, ..., config.num_labels - 1]`.

        Returns:

        """
        return_dict = return_dict if return_dict is not None else self.config.use_return_dict

        outputs = self.encoder(
            input_ids=input_ids,
            attention_mask=attention_mask,
            inputs_embeds=inputs_embeds,
            head_mask=head_mask,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
        )

        # Check encoder outputs for NaNs
        # self.check_for_nans(outputs[0], name="Encoder Output")
                              
        # outputs[0]: last_hidden_state
        sequence_output = outputs[0]
        sequence_output = self.dropout(sequence_output)

        '''
        # Check after dropout
        self.check_for_nans(sequence_output, name="After Dropout") 
        # Check for NaNs or Infs in classifier parameters
        for name, param in self.classifier.named_parameters():
            if torch.isnan(param).any():
                print(f"NaNs detected in classifier parameter: {name}")
            if torch.isinf(param).any():
                print(f"Infs detected in classifier parameter: {name}") 
        '''            
        logits = self.classifier(sequence_output)        

        # Check logits for NaNs
        # self.check_for_nans(logits, name="Logits")  

        loss = None
        if labels is not None:
            ignore_index = -1

            loss_fct = nn.CrossEntropyLoss(ignore_index=ignore_index)
            loss = loss_fct(logits.view(-1, self.num_labels), labels.view(-1))

        if not return_dict:
            output = (logits,) + outputs[2:]
            return ((loss,) + output) if loss is not None else output

        return TokenClassifierOutput(
            loss=loss,
            logits=logits,
            hidden_states=outputs.hidden_states,
            attentions=outputs.attentions
        )


import yaml
import torch
import csv
import transformers
from transformers import AutoTokenizer
from .T5_encoder import T5ForTokenClassification


class KSSDS:
    def __init__(self, config_path=None, model_path=None, tokenizer_path=None, max_repeats=60, detection_threshold=70, max_phrase_length=2):
        transformers.logging.set_verbosity_error()
        if config_path:
            with open(config_path, 'r') as file:
                self.config = yaml.safe_load(file)
        else:
            self.config = {
                "model_path": model_path or "ggomarobot/KSSDS",
                "tokenizer_path": tokenizer_path or "ggomarobot/KSSDS",
                "repetition_detection": {
                    "max_repeats": max_repeats,
                    "detection_threshold": detection_threshold,
                    "max_phrase_length": max_phrase_length
                },
                "max_length": 512,
            }

        self.model, self.device = self.load_model(self.config["model_path"])
        self.tokenizer = AutoTokenizer.from_pretrained(self.config["tokenizer_path"])
        self.max_length = self.config["max_length"]
        self.max_repeats = self.config["repetition_detection"]["max_repeats"]
        self.detection_threshold = self.config["repetition_detection"]["detection_threshold"]
        self.max_phrase_length = self.config["repetition_detection"]["max_phrase_length"]

    def load_model(self, model_path):
        model = T5ForTokenClassification.from_pretrained(model_path)
        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        model.to(device)
        return model, device

    def split_into_chunks(self, tokens):
        return [tokens[i:i + self.max_length] for i in range(0, len(tokens), self.max_length)]

    def process_text(self, text):
        tokens = self.tokenizer.encode(text, add_special_tokens=False)
        chunks = self.split_into_chunks(tokens)
        return chunks
    '''
    def handle_repetitions(self, text):
        words = text.split()
        if len(words) <= self.detection_threshold:
            return [text]

        result_sentences = []
        current_repetition = []
        current_sentence = []

        def flush_repetition():
            if len(current_repetition) > self.max_repeats:
                result_sentences.extend(
                    [" ".join(current_repetition[i:i + self.max_repeats]) for i in range(0, len(current_repetition), self.max_repeats)]
                )
            else:
                current_sentence.extend(current_repetition)

        for i, word in enumerate(words):
            if i > 0 and word == words[i - 1]:
                if current_sentence:
                    last_word = current_sentence.pop()
                    if current_sentence:
                        result_sentences.append(" ".join(current_sentence))
                    current_sentence = []
                    current_repetition.append(last_word)
                current_repetition.append(word)
            else:
                if current_repetition:
                    flush_repetition()
                    current_repetition = []
                current_sentence.append(word)

        if current_repetition:
            flush_repetition()
        if current_sentence:
            result_sentences.append(" ".join(current_sentence))

        return result_sentences
    '''
    def handle_repetitions(self, text):
        """
        Handles single-word and phrase repetitions in the text, ensuring proper order and separation.

        Args:
            text (str): The input text to process.

        Returns:
            List[str]: The processed text split into sentences.
        """
        words = text.split()
        if len(words) <= self.detection_threshold:
            return [text]

        result_sentences = []
        current_repetition = []
        current_sentence = []

        def flush_sentence():
            """Flush the current sentence into result_sentences."""
            if current_sentence:
                result_sentences.append(" ".join(current_sentence))
                current_sentence.clear()

        def flush_repetition():
            """Flush the current repetition into result_sentences."""
            for i in range(0, len(current_repetition), self.phrase_length * self.max_repeats):
                chunk = current_repetition[i:i + self.phrase_length * self.max_repeats]
                result_sentences.append(" ".join(chunk))
            current_repetition.clear()

        def find_repeating_phrase(start_idx):
            """Find the smallest repeating phrase starting at the given index."""
            for phrase_length in range(1, self.max_phrase_length + 1):
                phrase = words[start_idx:start_idx + phrase_length]
                next_idx = start_idx + phrase_length
                if next_idx + phrase_length <= len(words) and words[next_idx:next_idx + phrase_length] == phrase:
                    self.phrase_length = phrase_length  # Update `self.phrase_length` dynamically
                    return phrase
            return None

        i = 0
        self.phrase_length = 1  # Initialize with a default value
        while i < len(words):
            repeating_phrase = find_repeating_phrase(i)
            if repeating_phrase:
                # Flush any ongoing sentence before handling repetition
                flush_sentence()

                # Accumulate repeating phrases
                while i + self.phrase_length <= len(words) and words[i:i + self.phrase_length] == repeating_phrase:
                    current_repetition.extend(repeating_phrase)
                    i += self.phrase_length

                # Flush accumulated repetition if it reaches the threshold
                if len(current_repetition) >= self.phrase_length * self.max_repeats:
                    flush_repetition()
            else:
                # Add non-repeating words to the current sentence
                if current_repetition:
                    flush_repetition()
                current_sentence.append(words[i])
                i += 1

        # Flush any remaining tokens
        flush_sentence()
        flush_repetition()

        return result_sentences

        
    def segment_predictions(self, inp, pred):
        segments = []
        current_segment = []
        inp = inp[0]
        pred = pred[0]

        for token, label in zip(inp, pred):
            if label == 1:  # End of a sentence
                if current_segment:
                    current_segment.append(token)
                    segments.append(current_segment)
                    current_segment = []
                else:
                    segments.append([token])
            else:  # Continuation of a sentence
                current_segment.append(token)

        if current_segment:  # Add any remaining tokens as a final segment
            segments.append(current_segment)

        return segments


    def run_inference(self, input_sequence):
        chunks = self.process_text(input_sequence)
        results = []
        carry_over_tokens = []  # Tokens to carry over to the next chunk
        carry_over_labels = []  # Corresponding labels for carry-over tokens

        self.model.eval()
        with torch.inference_mode():
            for chunk in chunks:
                # Prepare inputs
                input_ids = torch.tensor([chunk]).to(self.device)
                attention_mask = torch.ones_like(input_ids).to(self.device)

                # Model inference
                outputs = self.model(input_ids=input_ids, attention_mask=attention_mask)
                predictions = torch.argmax(outputs.logits, dim=-1).squeeze().tolist()
                # Ensure predictions is a list, even for single-token inputs
                if isinstance(predictions, int):
                    predictions = [predictions]
                # Handle carry-over tokens from previous chunk
                if carry_over_tokens:
                    chunk = carry_over_tokens + chunk
                    predictions = carry_over_labels + predictions
                    carry_over_tokens = []
                    carry_over_labels = []
                # Segment predictions into sentences
                segmented_predictions = self.segment_predictions([chunk], [predictions])

                # Process each segment
                for i, segment in enumerate(segmented_predictions):
                    if i == len(segmented_predictions) - 1 and segment[-1] != 1:  # Last segment does not end in a sentence
                        carry_over_tokens = segment  # Carry over this segment
                        carry_over_labels = [0] * len(segment)  # Assign label 0 for carry-over tokens
                    else:
                        decoded_sentence = self.tokenizer.decode(segment, skip_special_tokens=False, clean_up_tokenization_spaces=False).strip()
                        # Handle repetitions
                        decoded_sentence = self.handle_repetitions(decoded_sentence)
                        if decoded_sentence:
                            results.extend(decoded_sentence)

        # Handle any remaining carry-over tokens
        if carry_over_tokens:
            decoded_remainder = self.tokenizer.decode(carry_over_tokens, skip_special_tokens=False, clean_up_tokenization_spaces=False).strip()
            decoded_remainder = self.handle_repetitions(decoded_remainder)
            if decoded_remainder:
                results.extend(decoded_remainder)

        return results

    # Wrapper function for KSSDS package; performs sentence splitting
    def split_sentences(self, input_sequence):
        """
        Split a single string input into sentences using the model.
        Args:
            input_sequence (str): The input text to be split.
        Returns:
            List[str]: A list of split sentences.
        """
        return self.run_inference(input_sequence)

    def process_tsv(self, input_tsv, output_tsv=None, output_print=False):
        # Get input column names from the configuration
        input_columns = self.config.get("input_columns", {})
        file_path_column = input_columns.get("file_path", "File Path")  # Default: "File Path"
        transcription_column = input_columns.get("transcription", "Transcription")  # Default: "Transcription"

        # Open the input TSV file
        with open(input_tsv, 'r', encoding='utf-8') as infile:
            reader = csv.DictReader(infile, delimiter='\t')
            fieldnames = ['File Name', 'Index', 'Sentence']  # Fixed output column names

            results = []

            # Process each row in the input TSV
            for row in reader:
                file_name = row.get(file_path_column, "").strip()  # Get file path
                transcription = row.get(transcription_column, "").strip()  # Get transcription

                # Split sentences using the KSSDS model
                split_sentences = self.run_inference(transcription)

                for idx, sentence in enumerate(split_sentences):
                    results.append((file_name, idx, sentence.strip()))

            # Write results to output TSV if specified
            if output_tsv:
                with open(output_tsv, 'w', encoding='utf-8', newline='') as outfile:
                    writer = csv.writer(outfile, delimiter='\t')
                    writer.writerow(fieldnames)
                    writer.writerows(results)

            # Print results to terminal if specified
            if output_print:
                for file_name, idx, sentence in results:
                    print(f"{file_name}\t{idx}\t{sentence}")

    def process_input_sequence(self, input_sequence, output_tsv=None, output_print=False):
        split_sentences = self.run_inference(input_sequence)

        if output_tsv:
            with open(output_tsv, 'w', encoding='utf-8', newline='') as outfile:
                writer = csv.writer(outfile, delimiter='\t')
                writer.writerow(['Index', 'Sentence'])
                for idx, sentence in enumerate(split_sentences):
                    writer.writerow([idx, sentence.strip()])

        if output_print:
            for idx, sentence in enumerate(split_sentences):
                print(f"[{idx}]: {sentence.strip()}")